#include "win.h"

#include <SDL.h>

#include <cctype>

const Uint32 MainWindow::flag_ = SDL_HWSURFACE|SDL_DOUBLEBUF;

Uint32 Window::getPixel(int x, int y) {
    int bpp = surf_->format->BytesPerPixel;
    /* Here p is the address to the pixel we want to retrieve */
    Uint8 *p = (Uint8 *)surf_->pixels + y * surf_->pitch + x * bpp;

    switch(bpp) {
    case 1:
        return *p;

    case 2:
        return *(Uint16 *)p;

    case 3:
        if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
            return p[0] << 16 | p[1] << 8 | p[2];
        else
            return p[0] | p[1] << 8 | p[2] << 16;

    case 4:
        return *(Uint32 *)p;

    default:
        return 0;       /* shouldn't happen, but avoids warnings */
    }

}

void Window::putPixel(int x, int y, Uint32 pixel) {
    int bpp = surf_->format->BytesPerPixel;
    /* Here p is the address to the pixel we want to set */
    Uint8 *p = (Uint8 *)surf_->pixels + y * surf_->pitch + x * bpp;

    switch(bpp) {
    case 1:
        *p = pixel;
        break;

    case 2:
        *(Uint16 *)p = pixel;
        break;

    case 3:
        if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
            p[0] = (pixel >> 16) & 0xff;
            p[1] = (pixel >> 8) & 0xff;
            p[2] = pixel & 0xff;
        } else {
            p[0] = pixel & 0xff;
            p[1] = (pixel >> 8) & 0xff;
            p[2] = (pixel >> 16) & 0xff;
        }
        break;

    case 4:
        *(Uint32 *)p = pixel;
        break;
    }

}

Window::Window(const Window& rhs)
	: size_(rhs.size()) {
	surf_ = SDL_ConvertSurface(rhs.surf_, rhs.surf_->format, SDL_SWSURFACE);
	assert(surf_ != 0);

	calcCache();
}

Window::Window(int x, int y, int bpp, Uint32 flag,
			   Uint32 Rmask, Uint32 Gmask, Uint32 Bmask, Uint32 Amask)
	: size_(x, y) {
    surf_ = SDL_CreateRGBSurface(flag, x, y, bpp, Rmask, Gmask, Bmask, Amask);
    assert(surf_ != 0);

	calcCache();
}

Window::Window(const Point& size, int bpp, Uint32 flag,
			   Uint32 Rmask, Uint32 Gmask, Uint32 Bmask, Uint32 Amask)
	: size_(size) {
	surf_ = SDL_CreateRGBSurface(flag, size.xi(), size.yi(), bpp,
								 Rmask, Gmask, Bmask, Amask);
	assert(surf_ != 0);

	calcCache();
}

void Window::printInformation(std::ostream& os) const {
    os << surf_->w << "*" << surf_->h << "*"
       << (int)surf_->format->BitsPerPixel << "\n";
}

void MainWindow::loadBMP(const std::string& str) {
    Window bw(str);
    blit(&bw);
}
void MainWindow::loadBMP(const std::string& str, const Point& pnt) {
    Window bw(str);

    Rect src(Point(0, 0), bw.size());
    Rect dst(pnt, Point(0, 0));
    blit(&bw, src, dst);
}

#ifdef HAVE_SDL_SDL_TTF_H

void MainWindow::putString(const std::string& str, const Point& pnt,
						   const Color& col) {
	Window text(TTF_RenderText_Solid(font_, str.c_str(), col.col()));
	blit(&text, pnt);
}

#else

const double MainWindow::dotPerChar_ = 300.0 / 41;
const int MainWindow::charWidth_ = 7;
const int MainWindow::charHeight_ = 14;

void MainWindow::putString(const std::string& str, const Point& pnt,
						   const Color&) {
    for (unsigned int i = 0; i < str.length(); i++) {
		char c = str[i];
		unsigned int n = sample_.find(std::toupper(c));
		if (n == std::string::npos) continue;

		Rect src(int(n*dotPerChar_), 0, charWidth_, charHeight_, true);
		Rect dst(pnt.xi()+int(i*dotPerChar_), pnt.yi(),
				 charWidth_, charHeight_, true);
		blit(strings_.get(), src, dst);
    }
}

#endif

void Window::convertDisplayFormat() {
	SDL_Surface* converted = SDL_DisplayFormat(surf_);
	assert(converted != 0);
	surf_ = converted;
}

void MainWindow::init(const std::string& str) {
    orig_.reset(new Window(str));
    first_.reset(new Window(str));
    reset();

#ifdef HAVE_SDL_SDL_TTF_H
	int ret = TTF_Init();
	assert(ret == 0);
	font_ = TTF_OpenFont("vixar.ttf", 16);
	assert(font_ != 0);
#else
	sample_ = std::string("0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ.<>^ ");
    strings_.reset(new Window("bmp/strings.bmp"));	
#endif
}

MainWindow::MainWindow(const std::string& str)
	: screen_(0) {
    surf_ = SDL_SetVideoMode(300, 400, 16, flag_);
    assert(surf_ != 0);	

	size_ = Point(300, 400);
	calcCache();

	init(str);
}

MainWindow::MainWindow(const std::string& str, int x, int y, Uint32 addFlag) {
    surf_ = SDL_CreateRGBSurface(SDL_SWSURFACE, 300, 400, 16, 0, 0, 0, 0);
    assert(surf_ != 0);

	screen_ = SDL_SetVideoMode(x, y, 16, flag_|addFlag);

	size_ = Point(x, y);
	calcCache();

	init(str);
}

MainWindow::~MainWindow() {
/*
#ifdef HAVE_SDL_SDL_TTF_H
	TTF_CloseFont(font_);
	TTF_Quit();
#endif
*/
}

void Window::setMask(Color col) {
	double rmask = static_cast<double>(col.r()) / 0xff;
	double gmask = static_cast<double>(col.g()) / 0xff;
	double bmask = static_cast<double>(col.b()) / 0xff;

	lock();
	for (int x = 0; x < surf_->w; x++) {
		for (int y = 0; y < surf_->h; y++) {
			Uint32 pixel = getPixel(x, y);
			if (pixel != surf_->format->colorkey) {
				Color before = getRGB(pixel);
				Color after(static_cast<int>(before.r()*rmask),
							static_cast<int>(before.g()*gmask),
							static_cast<int>(before.b()*bmask));
				putPixel(x, y, mapRGB(after));
			}
		}
	}
	unlock();
}

Window::Window(const std::string& str) {
    surf_ = SDL_LoadBMP(str.c_str());
    assert(surf_ != 0);

    size_ = Point(surf_->w, surf_->h);

	calcCache();
}

Window::~Window() {
    if (surf_ != 0) {
		SDL_FreeSurface(surf_);
	}
}

void Window::calcCache() {
	halfDiagonal_ = size_.length() / 2.0;
	halfSize_ = size_ / 2;
}

void MainWindow::clear() {
	blit(first_.get());
}

void MainWindow::clearRect(const Rect& rect) {
	blit(first_.get(), rect, rect);
}

void MainWindow::reset() {
	first_->blit(orig_.get());
	blit(first_.get());
}

void MainWindow::display() {
	if (screen_) {
		SDL_SoftStretch(surf_, 0, screen_, 0);
	}
}

int MainWindow::flip() {
	if (screen_) {
		display();
		return SDL_Flip(screen_);
	}
	else {
		return SDL_Flip(surf_);
	}
}
